==> raft-0.19.1.tar.gz: raft-0.19.1/src/membership.c <==
#include "membership.h"

#include "../include/raft.h"
#include "assert.h"
#include "configuration.h"
#include "err.h"
#include "heap.h"
#include "log.h"
#include "message.h"
#include "progress.h"
#include "queue.h"
#include "tracing.h"

#define infof(...) Infof(r->tracer, "  " __VA_ARGS__)

int membershipCanChangeConfiguration(struct raft *r)
{
    int rv;

    if (r->state != RAFT_LEADER || r->transfer != NULL) {
        rv = RAFT_NOTLEADER;
        goto err;
    }

    if (r->configuration_uncommitted_index != 0) {
        rv = RAFT_CANTCHANGE;
        goto err;
    }

    if (r->leader_state.promotee_id != 0) {
        rv = RAFT_CANTCHANGE;
        goto err;
    }

    /* In order to become leader at all we are supposed to have committed at
     * least the initial configuration at index 1. */
    assert(r->configuration_committed_index > 0);

    /* The index of the last committed configuration can't be greater than the
     * last log index. */
    assert(logLastIndex(r->log) >= r->configuration_committed_index);

    /* No catch-up round should be in progress. */
    assert(r->leader_state.round_number == 0);
    assert(r->leader_state.round_index == 0);
    assert(r->leader_state.round_start == 0);

    return 0;

err:
    assert(rv != 0);
    ErrMsgFromCode(r->errmsg, rv);
    return rv;
}

int membershipFetchLastCommittedConfiguration(struct raft *r,
                                              struct raft_configuration *conf)
{
    const struct raft_entry *entry;
    int rv;

    /* Try to get the entry at r->configuration_committed_index from the log. If
     * the entry is not present in the log anymore because the log was truncated
     * after a snapshot, we can just use configuration_last_snapshot, which we
     * cached when we took or restored the snapshot and is guaranteed to match
     * the content that the entry at r->configuration_committed_index had. */
    entry = logGet(r->log, r->configuration_committed_index);
    if (entry != NULL) {
        rv = configurationDecode(&entry->buf, conf);
    } else {
        assert(r->configuration_last_snapshot.n > 0);
        rv = configurationCopy(&r->configuration_last_snapshot, conf);
    }
    if (rv != 0) {
        return rv;
    }

    return 0;
}

bool membershipUpdateCatchUpRound(struct raft *r)
{
    unsigned server_index;
    raft_index match_index;
    raft_index last_index;
    raft_time round_duration;
    bool is_up_to_date;
    bool is_fast_enough;

    assert(r->state == RAFT_LEADER);
    assert(r->leader_state.promotee_id != 0);

    server_index =
        configurationIndexOf(&r->configuration, r->leader_state.promotee_id);
    assert(server_index < r->configuration.n);

    match_index = progressMatchIndex(r, server_index);

    /* If the server did not reach the target index for this round, it did not
     * catch up. */
    if (match_index < r->leader_state.round_index) {
        infof(
            "member (index: %u) not yet caught up match_index:%llu "
            "round_index:%llu",
            server_index, match_index, r->leader_state.round_index);
        return false;
    }

    last_index = logLastIndex(r->log);
    round_duration = r->now - r->leader_state.round_start;

    is_up_to_date = match_index == last_index;
    is_fast_enough = round_duration < r->election_timeout;

    infof("member is_up_to_date:%d is_fast_enough:%d", is_up_to_date,
          is_fast_enough);

    /* If the server's log is fully up-to-date or the round that just terminated
     * was fast enough, then the server as caught up. */
    if (is_up_to_date || is_fast_enough) {
        r->leader_state.round_number = 0;
        r->leader_state.round_index = 0;
        r->leader_state.round_start = 0;

        progressCatchUpFinish(r, server_index);

        return true;
    }

    /* If we get here it means that this catch-up round is complete, but there
     * are more entries to replicate, or it was not fast enough. Let's start a
     * new round. */
    r->leader_state.round_number++;
    r->leader_state.round_index = last_index;
    r->leader_state.round_start = r->now;

    return false;
}

int membershipUncommittedChange(struct raft *r,
                                const raft_index index,
                                const struct raft_entry *entry)
{
    struct raft_configuration configuration;
    int rv;

    assert(r != NULL);
    assert(r->state == RAFT_FOLLOWER);
    assert(entry != NULL);
    assert(entry->type == RAFT_CHANGE);

    rv = configurationDecode(&entry->buf, &configuration);
    if (rv != 0) {
        goto err;
    }

    raft_configuration_close(&r->configuration);

    r->configuration = configuration;
    r->configuration_uncommitted_index = index;

    return 0;

err:
    assert(rv != 0);
    return rv;
}

int membershipRollback(struct raft *r)
{
    int rv;

    assert(r != NULL);
    assert(r->state == RAFT_FOLLOWER);
    assert(r->configuration_uncommitted_index > 0);
    infof("roll back uncommitted configuration (%llu^%llu)",
          r->configuration_uncommitted_index,
          logTermOf(r->log, r->configuration_uncommitted_index));

    /* Fetch the last committed configuration entry. */
    assert(r->configuration_committed_index != 0);

    /* Replace the current configuration with the last committed one. */
    configurationClose(&r->configuration);
    rv = membershipFetchLastCommittedConfiguration(r, &r->configuration);
    if (rv != 0) {
        return rv;
    }

    r->configuration_uncommitted_index = 0;
    return 0;
}

void membershipLeadershipTransferInit(struct raft *r,
                                      struct raft_transfer *req,
                                      raft_id id,
                                      raft_transfer_cb cb)
{
    req->cb = cb;
    req->id = id;
    req->start = r->now;
    req->send.data = NULL;
    r->transfer = req;
}

int membershipLeadershipTransferStart(struct raft *r)
{
    const struct raft_server *server;
    struct raft_message message;
    int rv;
    assert(r->transfer->send.data == NULL);
    server = configurationGet(&r->configuration, r->transfer->id);
    assert(server != NULL);
    if (server == NULL) {
        return -1;
    }

    message.type = RAFT_IO_TIMEOUT_NOW;
    message.timeout_now.term = r->current_term;
    message.timeout_now.last_log_index = logLastIndex(r->log);
    message.timeout_now.last_log_term = logLastTerm(r->log);

    /* Set the data attribute of the raft_io_send object embedded in
     * raft_transfer. This is needed because we historically used it as a flag
     * to indicate that a transfer request was sent. See the replicationUpdate
     * function. */
    r->transfer->send.data = r;

    message.server_id = server->id;
    message.server_address = server->address;

    rv = MessageEnqueue(r, &message);
    if (rv != 0) {
        ErrMsgTransferf(r->io->errmsg, r->errmsg, "send timeout now to %llu",
                        server->id);
        return rv;
    }
    return 0;
}

void membershipLeadershipTransferClose(struct raft *r)
{
    struct raft_transfer *req = r->transfer;
    r->transfer = NULL;
    if (req->cb != NULL) {
        req->type = RAFT_TRANSFER_;
        QUEUE_PUSH(&r->legacy.requests, &req->queue);
    }
}

#undef infof

==> raft-0.21.0.tar.gz: raft-0.21.0/src/membership.c <==
#include "membership.h"

#include "../include/raft.h"
#include "assert.h"
#include "configuration.h"
#include "err.h"
#include "heap.h"
#include "message.h"
#include "progress.h"
#include "queue.h"
#include "tracing.h"
#include "trail.h"

#define infof(...) Infof(r->tracer, "  " __VA_ARGS__)

int membershipCanChangeConfiguration(struct raft *r)
{
    int rv;

    if (r->configuration_uncommitted_index != 0) {
        rv = RAFT_CANTCHANGE;
        goto err;
    }

    if (r->leader_state.promotee_id != 0) {
        rv = RAFT_CANTCHANGE;
        goto err;
    }

    /* In order to become leader at all we are supposed to have committed at
     * least the initial configuration at index 1. */
    assert(r->configuration_committed_index > 0);

    /* The index of the last committed configuration can't be greater than the
     * last log index. */
    assert(TrailLastIndex(&r->trail) >= r->configuration_committed_index);

    /* No catch-up round should be in progress. */
    assert(r->leader_state.round_number == 0);
    assert(r->leader_state.round_index == 0);
    assert(r->leader_state.round_start == 0);

    return 0;

err:
    assert(rv != 0);
    ErrMsgFromCode(r->errmsg, rv);
    return rv;
}

int membershipFetchLastCommittedConfiguration(struct raft *r,
                                              struct raft_configuration *conf)
{
    int rv;

    assert(r->configuration_committed_index > 0);
    rv = configurationCopy(&r->configuration_committed, conf);
    if (rv != 0) {
        return rv;
    }

    return 0;
}

bool membershipUpdateCatchUpRound(struct raft *r)
{
    unsigned server_index;
    raft_index match_index;
    raft_index last_index;
    raft_time round_duration;
    bool is_up_to_date;
    bool is_fast_enough;

    assert(r->state == RAFT_LEADER);
    assert(r->leader_state.promotee_id != 0);

    server_index =
        configurationIndexOf(&r->configuration, r->leader_state.promotee_id);
    assert(server_index < r->configuration.n);

    match_index = progressMatchIndex(r, server_index);

    /* If the server did not reach the target index for this round, it did not
     * catch up. */
    if (match_index < r->leader_state.round_index) {
        infof(
            "member (index: %u) not yet caught up match_index:%llu "
            "round_index:%llu",
            server_index, match_index, r->leader_state.round_index);
        return false;
    }

    last_index = TrailLastIndex(&r->trail);
    round_duration = r->now - r->leader_state.round_start;

    is_up_to_date = match_index == last_index;
    is_fast_enough = round_duration < r->election_timeout;

    infof("member is_up_to_date:%d is_fast_enough:%d", is_up_to_date,
          is_fast_enough);

    /* If the server's log is fully up-to-date or the round that just terminated
     * was fast enough, then the server as caught up. */
    if (is_up_to_date || is_fast_enough) {
        r->leader_state.round_number = 0;
        r->leader_state.round_index = 0;
        r->leader_state.round_start = 0;

        progressCatchUpFinish(r, server_index);

        return true;
    }

    /* If we get here it means that this catch-up round is complete, but there
     * are more entries to replicate, or it was not fast enough. Let's start a
     * new round. */
    r->leader_state.round_number++;
    r->leader_state.round_index = last_index;
    r->leader_state.round_start = r->now;

    return false;
}

int membershipUncommittedChange(struct raft *r,
                                const raft_index index,
                                const struct raft_entry *entry)
{
    struct raft_configuration configuration;
    int rv;

    assert(r != NULL);
    assert(r->state == RAFT_FOLLOWER);
    assert(entry != NULL);
    assert(entry->type == RAFT_CHANGE);

    rv = configurationDecode(&entry->buf, &configuration);
    if (rv != 0) {
        goto err;
    }

    raft_configuration_close(&r->configuration);

    r->configuration = configuration;
    r->configuration_uncommitted_index = index;

    return 0;

err:
    assert(rv != 0);
    return rv;
}

int membershipRollback(struct raft *r)
{
    int rv;

    assert(r != NULL);
    assert(r->state == RAFT_FOLLOWER);
    assert(r->configuration_uncommitted_index > 0);
    infof("roll back uncommitted configuration (%llu^%llu)",
          r->configuration_uncommitted_index,
          TrailTermOf(&r->trail, r->configuration_uncommitted_index));

    /* Fetch the last committed configuration entry. */
    assert(r->configuration_committed_index != 0);

    /* Replace the current configuration with the last committed one. */
    configurationClose(&r->configuration);
    rv = membershipFetchLastCommittedConfiguration(r, &r->configuration);
    if (rv != 0) {
        return rv;
    }

    r->configuration_uncommitted_index = 0;
    return 0;
}

int membershipLeadershipTransferStart(struct raft *r)
{
    const struct raft_server *server;
    struct raft_message message;
    int rv;

    assert(r->state == RAFT_LEADER);
    assert(r->leader_state.transferee != 0);
    assert(!r->leader_state.transferring);

    server = configurationGet(&r->configuration, r->leader_state.transferee);
    assert(server != NULL);

    if (server == NULL) {
        return -1;
    }

    message.type = RAFT_IO_TIMEOUT_NOW;
    message.timeout_now.version = MESSAGE__TIMEOUT_NOW_VERSION;
    message.timeout_now.term = r->current_term;
    message.timeout_now.last_log_index = TrailLastIndex(&r->trail);
    message.timeout_now.last_log_term = TrailLastTerm(&r->trail);

    message.server_id = server->id;
    message.server_address = server->address;

    infof("send timeout to %llu", server->id);
    rv = MessageEnqueue(r, &message);
    if (rv != 0) {
        ErrMsgPrintf(r->errmsg, "send timeout now to %llu", server->id);
        return rv;
    }

    /* Set the leadership transfer in progress flag. */
    r->leader_state.transferring = true;

    return 0;
}

#undef infof
